/*
 * Copyright (c)2004 The DragonFly Project.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 *   Neither the name of the DragonFly Project nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * fn_install.c
 * Installer Function : Install OS Files.
 * $Id: fn_install.c,v 1.34 2004/08/10 10:01:44 cpressey Exp $
 */

#include <string.h>

#include "aura/buffer.h"
#include "aura/fspred.h"

#include "dfui/dfui.h"

#include "installer/commands.h"
#include "installer/confed.h"
#include "installer/diskutil.h"
#include "installer/functions.h"
#include "installer/uiutil.h"

#include "flow.h"
#include "pathnames.h"

/*
 * Files to copy.
 * Split up the copy into small chunks; this lets us avoid copying all of
 * /usr/local, or other directories, if we so desire.
 */
char *cp_src[] = {
	"/COPYRIGHT",
	"/bin",
	"/boot",
	"/dev",
	"/etc",
	"/kernel",
	"/modules",
	"/root",
	"/sbin",
	"/sys",
	"/tmp",
	"/usr/bin",
	"/usr/games",
	"/usr/include",
	"/usr/lib",
#ifdef INSTALL_USR_LOCAL
	"/usr/local",
#endif
	"/usr/libdata",
	"/usr/libexec",
	"/usr/obj",
	"/usr/sbin",
	"/usr/share",
	"/usr/src",
	"/var",

	NULL
};

/*
 * fn_install_os: actually put DragonFly on a disk.
 */
void
fn_install_os(struct i_fn_args *a)
{
	struct subpartition *sp;
	struct commands *cmds;
	struct config_vars *rc_conf;
	char *string;
	int i, seen_it, prefix;

	cmds = commands_new();

	/*
	 * If swap isn't mounted yet, mount it.
	 */
	if (measure_activated_swap(a) == 0) {
		for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
		    sp != NULL; sp = subpartition_next(sp)) {
			if (!subpartition_is_swap(sp))
				continue;
			command_add(cmds, "%s%s %sdev/%ss%d%c",
			    a->os_root,
			    cmd_name(a, "SWAPON"),
			    a->os_root,
			    disk_get_device(storage_get_selected_disk(a->s)),
			    slice_get_number(storage_get_selected_slice(a->s)),
			    subpartition_get_letter(sp));
		}
	}

	/*
	 * Unmount anything already mounted on /mnt.
	 */
	unmount_all_under(a, cmds, "%smnt", a->os_root);

	/*
	 * Create mount points and mount subpartitions on them.
	 */
	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_next(sp)) {
		if (subpartition_is_swap(sp)) {
			/*
			 * Set this subpartition as the dump device.
			 */
#ifdef AUTOMATICALLY_ENABLE_CRASH_DUMPS
			if (subpartition_get_capacity(sp) < storage_get_memsize(a->s))
				continue;

			command_add(cmds, "%s%s -v %sdev/%ss%d%c",
			    a->os_root, cmd_name(a, "DUMPON"),
			    a->os_root,
			    disk_get_device(storage_get_selected_disk(a->s)),
			    slice_get_number(storage_get_selected_slice(a->s)),
			    subpartition_get_letter(sp));

			rc_conf = config_vars_new();
			asprintf(&string, "/dev/%ss%d%c",
			    disk_get_device(storage_get_selected_disk(a->s)),
			    slice_get_number(storage_get_selected_slice(a->s)),
			    subpartition_get_letter(sp));
			config_var_set(rc_conf, "dumpdev", string);
			free(string);
			config_var_set(rc_conf, "dumpdir", "/var/crash");
			config_vars_write(rc_conf, CONFIG_TYPE_SH,
			    "%s%setc/rc.conf", a->os_root, "/");
			config_vars_free(rc_conf);
#endif
			continue;
		}
	
		if (strcmp(subpartition_get_mountpoint(sp), "/") != 0) {
			command_add(cmds, "%s%s -p %smnt%s",
			    a->os_root, cmd_name(a, "MKDIR"),
			    a->os_root, subpartition_get_mountpoint(sp));
		}

		command_add(cmds, "%s%s %sdev/%ss%d%c %smnt%s",
		    a->os_root, cmd_name(a, "MOUNT"),
		    a->os_root,
		    disk_get_device(storage_get_selected_disk(a->s)),
		    slice_get_number(storage_get_selected_slice(a->s)),
		    subpartition_get_letter(sp),
		    a->os_root,
		    subpartition_get_mountpoint(sp));
	}

	/*
	 * Actually copy files now.
	 */

	for (i = 0; cp_src[i] != NULL; i++) {
		char *dir = cp_src[i];

		command_add_logless(cmds, "%s%s %s%s %smnt%s",
		    a->os_root, cmd_name(a, "CPDUP"),
		    a->os_root, &dir[1],
		    a->os_root, dir);
	}

	/*
	 * Now, because cpdup does not cross mount points,
	 * we must copy anything that the user might've made a
	 * seperate mount point for (e.g. /usr/libdata/lint.)
	 */
	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_next(sp)) {
		/*
		 * If the subpartition is a swap subpartition, don't
		 * try to copy anything into it.
		 */
		if (subpartition_is_swap(sp))
			continue;

		/*
		 * If the mountpoint doesn't even exist on the installation
		 * medium, don't try to copy anything from it!  We assume
		 * it's an empty subpartition for the user's needs.
		 */
		if (!is_dir("%s%s", a->os_root, &subpartition_get_mountpoint(sp)[1]))
			continue;

		/*
		 * Don't bother copying the mountpoint IF:
		 * - we've already said to copy it, or something besides it
		 *   (it's a prefix of something in cp_src); or
		 * - we haven't said to copy it
		 *   (nothing in cp_src is a prefix of it.)
		 */
		seen_it = 0;
		prefix = 0;
		for (i = 0; cp_src[i] != NULL; i++) {
			if (strncmp(subpartition_get_mountpoint(sp), cp_src[i],
			    strlen(subpartition_get_mountpoint(sp))) == 0) {
				seen_it = 1;
				break;
			}
			if (strncmp(cp_src[i], subpartition_get_mountpoint(sp),
			    strlen(cp_src[i])) == 0) {
				prefix = 1;
			}
		}
		if (seen_it || !prefix)
			continue;

		/*
		 * Otherwise, cpdup the subpartition.
		 */
		command_add_logless(cmds, "%s%s %s%s %smnt%s",
		    a->os_root, cmd_name(a, "CPDUP"),
		    a->os_root, &subpartition_get_mountpoint(sp)[1],
		    a->os_root, subpartition_get_mountpoint(sp));
	}
	
	/*
	 * Create symlinks.
	 */

	/*
	 * If the user has both /var and /tmp subparitions,
	 * symlink /var/tmp to /tmp.
	 */
	if (subpartition_find(storage_get_selected_slice(a->s), "/tmp") != NULL &&
	    subpartition_find(storage_get_selected_slice(a->s), "/var") != NULL) {
		command_add(cmds, "%s%s 1777 %smnt/tmp",
		    a->os_root, cmd_name(a, "CHMOD"), a->os_root);
		command_add(cmds, "%s%s -rf %smnt/var/tmp",
		    a->os_root, cmd_name(a, "RM"), a->os_root);
		command_add(cmds, "%s%s -s /tmp %smnt/var/tmp",
		    a->os_root, cmd_name(a, "LN"), a->os_root);
	}

	/*
	 * If the user has /var, but no /tmp,
	 * symlink /tmp to /var/tmp.
	 */
	if (subpartition_find(storage_get_selected_slice(a->s), "/tmp") == NULL &&
	    subpartition_find(storage_get_selected_slice(a->s), "/var") != NULL) {
		command_add(cmds, "%s%s -rf %smnt/tmp",
		    a->os_root, cmd_name(a, "RM"), a->os_root);
		command_add(cmds, "%s%s -s /var/tmp %smnt/tmp",
		    a->os_root, cmd_name(a, "LN"), a->os_root);
	}

	/*
	 * If the user has /usr, but no /home,
	 * symlink /home to /usr/home.
	 */
	if (subpartition_find(storage_get_selected_slice(a->s), "/home") == NULL &&
	    subpartition_find(storage_get_selected_slice(a->s), "/usr") != NULL) {
		command_add(cmds, "%s%s -rf %smnt/home",
		    a->os_root, cmd_name(a, "RM"), a->os_root);
		command_add(cmds, "%s%s -s /usr/home %smnt/home",
		    a->os_root, cmd_name(a, "LN"), a->os_root);
	}

	/*
	 * XXX check for other possible combinations too?
	 */

	/*
	 * Clean up.  In case some file didn't make it, use rm -f
	 */
	command_add(cmds, "%s%s -f %smnt/boot/loader.conf",
	    a->os_root, cmd_name(a, "RM"), a->os_root);
	command_add(cmds, "%s%s -f %smnt/etc/rc.d/pfi",
	    a->os_root, cmd_name(a, "RM"), a->os_root);
	command_add(cmds, "%s%s -f %smnt/etc/issue",
	    a->os_root, cmd_name(a, "RM"), a->os_root);
	command_add(cmds, "%s%s -f %smnt/tmp/install.log",
	    a->os_root, cmd_name(a, "RM"), a->os_root);

	/*
	 * Copy pristine versions of any files we might have installed
	 * that only have applicability to Live CD boots.
	 */
	command_add_logless(cmds,
	    "%s%s %susr/local/share/pristine/etc %smnt/etc",
	    a->os_root, cmd_name(a, "CPDUP"), a->os_root, a->os_root);

	/*
	 * Rebuild the user database, to get rid of any extra users
	 * from the LiveCD that aren't supposed to be installed
	 * (copying a pristine master.passwd isn't enough.)
	 */
	command_add(cmds, "%s%s -p -d %smnt/etc %smnt/etc/master.passwd",
	    a->os_root, cmd_name(a, "PWD_MKDB"), a->os_root, a->os_root);

	/* Create missing directories. */
	command_add(cmds, "%s%s %smnt/proc",
	    a->os_root, cmd_name(a, "MKDIR"), a->os_root);
	command_add(cmds, "%s%s %smnt/mnt",
	    a->os_root, cmd_name(a, "MKDIR"), a->os_root);

	/* Write new fstab. */

	command_add(cmds, "%s%s '%s' >%smnt/etc/fstab",
	    a->os_root, cmd_name(a, "ECHO"),
	    "# Device\t\tMountpoint\tFStype\tOptions\t\tDump\tPass#",
	    a->os_root);

	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_next(sp)) {
		if (strcmp(subpartition_get_mountpoint(sp), "/") == 0) {
			command_add(cmds, "%s%s '/dev/%ss%d%c\t\t%s\t\tufs\trw\t\t1\t1' >>%smnt/etc/fstab",
			    a->os_root, cmd_name(a, "ECHO"),
			    disk_get_device(storage_get_selected_disk(a->s)),
			    slice_get_number(storage_get_selected_slice(a->s)),
			    subpartition_get_letter(sp),
			    subpartition_get_mountpoint(sp),
			    a->os_root);
		} else if (strcmp(subpartition_get_mountpoint(sp), "swap") == 0) {
			command_add(cmds, "%s%s '/dev/%ss%d%c\t\tnone\t\tswap\tsw\t\t0\t0' >>%smnt/etc/fstab",
			    a->os_root, cmd_name(a, "ECHO"),
			    disk_get_device(storage_get_selected_disk(a->s)),
			    slice_get_number(storage_get_selected_slice(a->s)),
			    subpartition_get_letter(sp),
			    a->os_root);
		} else {
			command_add(cmds, "%s%s '/dev/%ss%d%c\t\t%s\t\tufs\trw\t\t2\t2' >>%smnt/etc/fstab",
			    a->os_root, cmd_name(a, "ECHO"),
			    disk_get_device(storage_get_selected_disk(a->s)),
			    slice_get_number(storage_get_selected_slice(a->s)),
			    subpartition_get_letter(sp),
			    subpartition_get_mountpoint(sp),
			    a->os_root);
		}
	}

	command_add(cmds, "%s%s '%s' >>%smnt/etc/fstab",
	    a->os_root, cmd_name(a, "ECHO"),
	    "proc\t\t\t/proc\t\tprocfs\trw\t\t0\t0",
	    a->os_root);

	/* Backup the disklabel and the log. */

	command_add(cmds, "%s%s %ss%d > %smnt/etc/disklabel.%ss%d",
	    a->os_root, cmd_name(a, "DISKLABEL"),
	    disk_get_device(storage_get_selected_disk(a->s)),
	    slice_get_number(storage_get_selected_slice(a->s)),
	    a->os_root,
	    disk_get_device(storage_get_selected_disk(a->s)),
	    slice_get_number(storage_get_selected_slice(a->s)));
	command_add(cmds, "%s%s %sinstall.log %smnt/var/log/install.log",
	    a->os_root, cmd_name(a, "CP"),
	    a->tmp, a->os_root);
	command_add(cmds, "%s%s 600 %smnt/var/log/install.log",
	    a->os_root, cmd_name(a, "CHMOD"), a->os_root);

	/*
	 * Do it!
	 */
	/* commands_preview(cmds); */
	if (!commands_execute(a, cmds)) {
		inform(a->c, "DragonFly was not fully installed.");
		a->result = 0;
	} else {
		a->result = 1;
	}
	commands_free(cmds);

	/*
	 * Unmount everything we mounted on /mnt.  This is done in a seperate
	 * command chain, so that partitions are unmounted, even if an error
	 * occurs in one of the preceding commands, or it is cancelled.
	 */
	cmds = commands_new();
	unmount_all_under(a, cmds, "%smnt", a->os_root);

	/*
	 * Once everything is unmounted, if the install went successfully,
	 * make sure once and for all that the disklabel is bootable.
	 */
	if (a->result) {
		command_add(cmds, "%s%s -B %ss%d",
		    a->os_root, cmd_name(a, "DISKLABEL"),
		    disk_get_device(storage_get_selected_disk(a->s)),
		    slice_get_number(storage_get_selected_slice(a->s)));
	}

	if (!commands_execute(a, cmds))
		inform(a->c, "Warning: subpartitions were not correctly unmounted.");

	commands_free(cmds);
}
